program heat_fem2d_triangle

	! This program is designed to solve the approximation to the function:
	! qq*ut - pp\Delta u = ff(x)
    ! or
    ! qq*ut - pp(uxx + uyy) = ff(x)
	
	implicit none
	
	! Lets Define the variables we are going to need:
		
	! The dimension we are in
	integer (kind=4), parameter :: xdim = 2
	! A parameter to allocate the arrays
    integer (kind=4), parameter :: n_alloc = 25000
    ! Dirichlet(1) or Neumann BCs(0)
	integer (kind=4), parameter :: BCb = 1
	! The left and right boundary conditions
	real (kind=8), parameter :: ub = 0
    ! The total time that we want to step to
    real (kind=8), parameter :: T = 1.0
	! The number of quadrature points to use for the integration
	integer (kind=4), parameter :: nquad = 1
	! The types of basis functions to use per interval (4=bilinear) (3= simplex)
	integer (kind=4), parameter :: nb = 3
    ! The filename that the solution will be written to
    character (len=80) :: SOLUTION_csv_file_name = 'Solution.csv'
	
	
	! These variables we will read in from the geometry routine
	
	! total number of points
	integer (kind=4) n
	! number of elements
	integer (kind=4) nelements
	! This array stores the x values at the nodes
	real (kind=8) x(1:xdim,1:n_alloc)
	! This array stores the quadrature points
	real (kind=8) xquad(1:xdim,1:n_alloc)
	! The discretization we get
	real (kind=8) h
	! The time discretization
    real (kind=8) dt 
	! The number of time steps we need to take
    integer (kind=4) nt
	! This array keeps track of the global and local nodes
	! at each of our elements
	integer (kind=4) lgnode(1:nb,1:n_alloc)
	! This value keeps track of how many unknowns we have
	integer (kind=4) nunk
	! This array keeps track of the unknown nodes
	integer (kind=4) unknode(1:n_alloc)

	
	! These are variables that we will be creating/editing
	
    ! The current time
    real (kind=8) time
	! The bands of the matrix will be stored in columns of our
	! matrix A, initialize to be size n
	real (kind=8) init_A(1:n_alloc,1:n_alloc)
	real (kind=8) A(1:n_alloc,1:n_alloc)
    ! This is the pivot array that we may get from the lapack solver
    real (kind=8) IPIV(1:n_alloc)
	! The right hand side of our linear system
	real (kind=8) init_rhs(1:n_alloc)
	real (kind=8) rhs(1:n_alloc)
    ! The solution at different time steps
    real (kind=8) u(1:n_alloc)
	! The Error
	real (kind=8) L2Err
	real (kind=8) H1Err
	
	! These are just extra variables that we may need
	integer (kind=4) i
	integer (kind=4) j
    integer (kind=4) it
	integer (kind=4) iUj
    integer (kind=4) iunk
	integer (kind=4) sucess !Whether the solve suceeded or not
    integer (kind=4) csv_file_unit
    character (len=n_alloc) record
	
	
	! Define the Geometry of the problem
	call geometry(xdim, nquad, nb, &
                  n, nelements, x, xquad, h,&
                  lgnode, nunk, unknode)
	
	! Set the temporal discretization
	dt = h**2
	nt = T/dt
	
	! Use this for printing things
	write(*,*) 'The Following is output from the Geometry Routine:'
	write(*,'(A,i2)') 'Number of nodes:', n
	write(*,'(A8,A8,A16)') 'node','x','y'
	do i = 1, n
		write ( *, '(2x,i8,2x,g14.6,2x,g14.6)' )  i, x(1,i), x(2,i)
	end do
	
	write(*,*) 'The quadrature points at each of our elements:'
    write(*,'(A8,A8,A16)') 'node','x','y'
	do i = 1, nelements
		write ( *, '(2x,i8,2x,g14.6,2x,g14.6)' )  i, xquad(1,i), xquad(2,i)
	end do
	 
	write(*,*) 'The local/Global node array'
	write(*,'(A10,A16,A16,A16)') 'Element', 'Node 1', 'Node 2','Node 3'
	do i = 1, nelements
		write ( *, '(2x,i8,2x,g14.6,2x,g14.6,2x,g14.6)' )  i, &
                                 lgnode(1,i), lgnode(2,i), lgnode(3,i)
	end do
    

	write(*,*) 'The unknown number for the corresponding node:'
	do i = 1, n
		write ( *, '(2x,i8,2x,g14.6)' )  i, unknode(i)
	end do
        
    call csv_file_open_write(SOLUTION_csv_file_name, csv_file_unit)
    
    ! Initialize the solution vector for the previous time step
    u(1:n) = 0.0
    
    record = ''
    
    ! Write the initial solution to the file
    ! Using homogeneous dirichlet BC's
    do i = 1, n 
        ! At a boundary
        if(unknode(i) == -1) then 
            call csv_record_append_r8(ub, record)
            
        else
            call csv_record_append_r8(u(i), record)
        end if
    end do
    
    call csv_file_record_write(SOLUTION_csv_file_name, csv_file_unit, record)
       

    !
    ! Before we enter the time loop we want to assemble the A matrix
    ! because that will not change as we change our time
    !
    
    call assemble_A(nelements,n,nquad, &
                    xdim,x(1:xdim,1:n),xquad(1:xdim,1:nelements), &
                    BCb,ub, &
                    nb,lgnode(1:nb,1:nelements),unknode(1:n),nunk, &
                    dt, &
                    init_A(1:nunk,1:nunk),init_rhs(1:nunk))
    
	! Append the errors to a file
	! Open a file
	!open (unit = 8, file = "A.txt")
	
    ! Debugging
    !do i = 1, nunk
    !    do j = 1, nunk
    !        write(*,'(2x,g14.6)',advance='NO') init_A(i,j)
    !    end do
    !    write(*,*)
     !   write(*,'(2x,g14.6)') init_rhs(i)
    !end do
	!close (8)
     
    ! Initially the time is zero
    time = 0.0
    
    ! Now we need to solve the problem at each
    do it = 1, nt
        
        ! Time at this particular time step
        time = time + dt

        ! Set the new rhs to what we got from the assemble A matrix
        rhs(1:nunk) = init_rhs
        A(1:nunk,1:nunk) = init_A(1:nunk,1:nunk)
        
        ! This assemble routine will put together the matrix that is our
        ! left hand side
        call assemble_rhs(nelements,n,nquad, &
                          xdim,x(1:xdim,1:n),xquad(1:xdim,1:nelements),&
                          BCb,ub, &
                          nb,lgnode(1:nb,1:nelements),unknode(1:n),nunk, &
                          dt,time,u(1:n), &
                          rhs(1:nunk))
        !do i = 1, nunk
        !    write(*,*) rhs(i)
        !end do

        
        IPIV(1:nunk) = 0.0
        
        ! Use a general lapack solver routine, If time permits a banded solver will be investigated
        call dgesv(nunk, 1, A(1:nunk,1:nunk), nunk, IPIV(1:nunk), rhs(1:nunk), nunk, sucess) 
        ! For debugging:

        write(*,*) sucess
        		
		! set the counters
		iunk = 0
		
		! Set the solution for the next time step
        record = ''
    
        ! Write the initial solution to the file
        ! Using homogeneous dirichlet BC's
        do i = 1, n 
            ! At a boundary
            if(unknode(i) == -1) then 
                call csv_record_append_r8(ub, record)
				u(i) = ub
            else
                iunk = iunk + 1 
                u(i) = rhs(iunk)
                call csv_record_append_r8(u(i), record)
            end if
        end do
        
    
        call csv_file_record_write(SOLUTION_csv_file_name, csv_file_unit, record)
        
                       
        ! Add a new line
        write(8,*)
        
        ! Call the error routine so that we  can determine the error and stuff
        !call error (nelements,n,nb,nunk, &
        !            xdim,x, &
        !            lgnode,unknode, &
        !            dt,time, &
        !            u, &
        !            L2Err,H1Err)
        
        !write(*,*) L2Err
        !write(*,*) H1Err
       
        
    end do ! End the loop over the time steps
    
    ! Close the solution file
    call csv_file_close_write(SOLUTION_csv_file_name, csv_file_unit)
	write(*,*) dt
	
    ! Open files for the the geometry
	open (unit = 9, file = "geometry.txt")
    do i = 1, n
        write(9,'(2x,g14.6,2x,g14.6)') x(1,i), x(2,i)
    end do
    close(9)
    
	time = 1.0
	! Call the error routine so that we  can determine the error and stuff
	call error(nelements, n, nb, nunk, &
               xdim, x, &
		       lgnode, unknode, &
               dt, time, &
			   u, &
			   L2Err, H1Err)
	
    write(*,*) L2Err
    write(*,*) H1Err
    
	! Append the errors to a file
	! Open a file
	open (unit = 8, file = "Errors.txt", access = "APPEND")
	
	! Append the errors
	write(8,'(2x,f16.10,2x,f16.10,2x,f16.10)') h, L2Err, H1Err

	! Close the file
	close(8)
	
	
	! Now write out the solution vector to a file so that we can read it into
	! matlab or gnuplot and make some pretty plots
	
	! Open a file
	!open (unit = 9, file = "solution.txt")
	
	!do i = 1, n
	!	write(9,'(2x,f16.10,2x,f16.10)') x(i),uj(i)
		!write(*,'(f16.10)') rhs(i)
	!close(9)
	
	
end

subroutine assemble_A(nelements,n,nquad, &
                      xdim,x,xquad, &
                      BCb,ub, &
                      nb,lgnode,unknode,nunk, &
                      dt, &
                      A,rhs)
	! This routine will put together the matrix we must use to solve
	! The linear system
	
	implicit none
	
	! Inputs
	
	! Info about the grid
	integer (kind=4) nelements
	integer (kind=4) n
	integer (kind=4) nquad
    
    ! The xy coordinates and temporal discretization
    integer (kind=4) xdim
	real (kind=8) x(1:xdim,1:n)
	real (kind=8) xquad(1:xdim,1:nelements)
	
	! The Boundary conditions
	real (kind=8) xb
	! What types of Boundary conditions we are dealing with
	! Dirichlet = 1
	! Neumann = 0
	integer (kind=4) BCb
	real (kind=8) ub

	! The nodal lookup tables and basis functions we will use
    integer (kind=4) nb
	integer (kind=4) lgnode(1:nb,1:nelements)
	integer (kind=4) unknode(1:n)
	integer (kind=4) nunk
	
    ! The solution vector from the previous time step
    real (kind=8) dt
    real (kind=8) time
    
	! Outputs
	
	! The bands of the matrix
	real (kind=8) A(1:nunk,1:nunk)
    ! The right hand side
    real (kind=8) rhs(1:nunk)
	
	! Subroutine specific variables
	
	! The local x positions of the nodes
	real (kind=8) lx1(1:xdim)
	real (kind=8) lx2(1:xdim)
	real (kind=8) lx3(1:xdim)
    ! Squared edge lengths of the triangle
    real (kind=8) b
    real (kind=8) c
    real (kind=8) d
	! The quadrature weight
	real (kind=8) w
	! The coefficient and rhs values to our DE
	real (kind=8) pp
	real (kind=8) qq
	! Values of the basis functions
	real (kind=8) phii
	real (kind=8) phiix
	real (kind=8) phiiy
	real (kind=8) phij
	real (kind=8) phijx
	real (kind=8) phijy
	! Contribution to the inner products
	real (kind=8) aij
	! Counters
	integer (kind=4) iE !elements
	integer (kind=4) iQ !quadrature points
	integer (kind=4) iB !Basis function i
	integer (kind=4) jB !Basis function j
	integer (kind=4) iglobal ! global node 
	integer (kind=4) jglobal ! global node we are testing
	integer (kind=4) iunk !unknown position we are on
	integer (kind=4) junk !unknown position we are testing
    
    ! For debuging
    integer (kind=4) i, j
        
	! Initialize the values of the matrix to be zero
	A(1:nunk,1:nunk) = 0
    rhs(1:nunk) = 0
      
	!
	! This is the heart of the Assembly
	
	! For all of the elements
	do iE = 1, nelements
		
		! Get the size of the interval based on the left
		! and right nodal values
        ! starts in bottom left then goes counter-clockwise
		lx1 = x(:,lgnode(1,iE))
		lx2 = x(:,lgnode(2,iE))
		lx3 = x(:,lgnode(3,iE))
        
        ! Calculate the area of the triangle we are looking at using Herons formula
        ! This is a modification of the formula in order to avoid doing multiple
        ! square roots
        b = ((lx2(2)-lx1(2))**2) + ((lx2(1)-lx1(1))**2)
        c = ((lx3(2)-lx2(2))**2) + ((lx3(1)-lx2(1))**2)
        d = ((lx1(2)-lx3(2))**2) + ((lx1(1)-lx3(1))**2)
        w = (1.0/4.0)*sqrt(abs(((b + c + d)**2) - &
		    (2*((b**2)+(c**2)+(d**2)))))           
	
		! Loop over the quadrature points
		do iQ = 1, nquad
			
			! Loop over the basis functions on our element
			do iB = 1, nb
				
				! Get the global node and then the
				! unknown position we are looking at
				iglobal = lgnode(iB,iE)
				iunk = unknode(iglobal)
				
				! Check to make sure that we are not at a Dirichlet BC
				if(iunk > 0) then
			
					! Find the value of phi_i and phix_i
					call phi(xdim, xquad(:,iE), iB, lx1, lx2, lx3, phii, phiix, phiiy)
					                    
					
					! Now loop over the other basis functions
					do jB = 1, nb
						
						! Get the Global and unknown position we are
						! testing against
						jglobal = lgnode(jB,iE)
						junk = unknode(jglobal)

						! Get the test function
						call phi(xdim, xquad(:,iE), jB, lx1, lx2, lx3, phij, phijx, phijy)
						
						! Find the contribution aij
						aij = w*(pp(xdim,xquad(:,iE))*(phiix*phijx + phiiy*phijy) + &
                                ((1/dt)*qq(xdim,xquad(:,iE))*phii*phij))
                                                      
                        ! write(*,'(2x,2x,g14.6,2x,2x,g14.6,2x,2x,g14.6)') lxx1, lxx2,aij
						
						! If the unknown test function we are at is a known BC
						! then we can add it to the right hand side
						if(junk < 0) then
							
							! If we are at a boundary
							if(junk == -1) then
								rhs(iunk) = rhs(iunk) - aij*ub
							end if
						
						! Otherwise then we can put it in the matrix 
						else
                            
                            ! Not the most efficient storage.  But for small problems,
                            ! this is acceptable
                            A(iunk,junk) = A(iunk,junk) + aij
                            
						end if ! if we are/arent at an unknown node
					end do ! Loop over the test functions
				end if ! If we are not at a Dirichlet BC				
			end do ! Loop over the trial functions
		end do ! Loop over the quadrature points
	end do ! Loop over the elements		
	
end

subroutine assemble_rhs(nelements,n,nquad, &
                    xdim,x,xquad,&
                    BCb,ub, &
                    nb,lgnode,unknode,nunk, &
                    dt,time,ukm1, &
                    rhs)
	! This routine will put together the matrix we must use to solve
	! The linear system
	
	implicit none
	
	! Inputs
	
	! Info about the grid
	integer (kind=4) nelements
	integer (kind=4) n
	integer (kind=4) nquad
    
    ! The xy coordinates and temporal discretization
    integer (kind=4) xdim
	real (kind=8) x(1:xdim,1:n)
	real (kind=8) xquad(1:xdim,1:nelements)
	
	! What types of Boundary conditions we are dealing with
	! Dirichlet = 1
	! Neumann = 0
	integer (kind=4) BCb
	real (kind=8) ub
	
	! The nodal lookup tables and basis functions we will use
    integer (kind=4) nb
	integer (kind=4) lgnode(1:nb,1:nelements)
	integer (kind=4) unknode(1:n)
	integer (kind=4) nunk
	
    ! The solution vector from the previous time step
    real (kind=8) dt
    real (kind=8) time
    real (kind=8) ukm1(1:n)
    
	! Outputs
	
	! The right hand side
	real (kind=8) rhs(1:nunk)
	
	! Subroutine specific variables
	
	! The local x positions of the nodes
	real (kind=8) lx1(1:xdim)
	real (kind=8) lx2(1:xdim)
	real (kind=8) lx3(1:xdim)
    ! Squared edge lengths of the triangle
    real (kind=8) b
    real (kind=8) c
    real (kind=8) d
	! The quadrature weight
	real (kind=8) w
	! The coefficient and rhs values to our DE
	real (kind=8) ff
	real (kind=8) pp
	real (kind=8) qq
	! Values of the basis functions
	real (kind=8) phii
	real (kind=8) phiix
	real (kind=8) phiiy
	real (kind=8) phij
	real (kind=8) phijx
	real (kind=8) phijy
	! ukm1 at quadrature point
	real (kind=8) ukm1_qp
	! Counters
	integer (kind=4) iE !elements
	integer (kind=4) iQ !quadrature points
	integer (kind=4) iB !Basis function i
	integer (kind=4) jB !Basis function j
	integer (kind=4) iglobal ! global node 
	integer (kind=4) jglobal ! global node we are testing
	integer (kind=4) iunk !unknown position we are on
	integer (kind=4) junk !unknown position we are testing
    
    ! For debuging
    integer (kind=4) i, j
                
	!
	! This is the heart of the Assembly

	! For all of the elements
	do iE = 1, nelements
		
		! Get the size of the interval based on the left
		! and right nodal values
        ! starts in bottom left then goes counter-clockwise
		lx1 = x(:,lgnode(1,iE))
		lx2 = x(:,lgnode(2,iE))
		lx3 = x(:,lgnode(3,iE))
	
        ! Calculate the area of the triangle we are looking at using Herons formula
        ! This is a modification of the formula in order to avoid doing multiple
        ! square roots
        b = ((lx2(2)-lx1(2))**2) + ((lx2(1)-lx1(1))**2)
        c = ((lx3(2)-lx2(2))**2) + ((lx3(1)-lx2(1))**2)
        d = ((lx1(2)-lx3(2))**2) + ((lx1(1)-lx3(1))**2)
        w = (1.0/4.0)*sqrt(abs(((b + c + d)**2) - &
		    (2*((b**2)+(c**2)+(d**2))))) 
    
		! Loop over the quadrature points
		do iQ = 1, nquad
			
			
			! Loop over the basis functions on our element
			do iB = 1, nb
				
				! Get the global node and then the
				! unknown position we are looking at
				iglobal = lgnode(iB,iE)
				iunk = unknode(iglobal)
				
				! Check to make sure that we are not at a Dirichlet BC
				if(iunk > 0) then
			
					! Find the value of phi_i and phix_i
					call phi(xdim, xquad(:,iE), iB, lx1, lx2, lx3, phii, phiix, phiiy)
					                    
                    !write(*,*) phii
					
					! Find the value of ukm1 at the quadrature point
					ukm1_qp = (ukm1(lgnode(1,iE)) + ukm1(lgnode(2,iE)) + &
                               ukm1(lgnode(3,iE)))/3.0
					
					! Update the right hand side
					rhs(iunk) = rhs(iunk) + w*(ff(xdim,xquad(:,iE),time)*phii + &
                                (1/dt)*qq(xdim,xquad(:,iE))*ukm1_qp*phii)
					
				end if ! If we are not at a Dirichlet BC				
			end do ! Loop over the trial functions
		end do ! Loop over the quadrature points
	end do ! Loop over the elements		
	
end

subroutine error(nelements, n, nb, nunk, &
                 xdim, x, &
		         lgnode, unknode, &
                 dt, time, &
				 u, &
				 L2Err, H1Err)
	! This error routine will calculate the L2 error based on our
	! solution to the system of Linear equations
	
	implicit none
	
	! The inputs we need
	integer (kind=4) nelements
	integer (kind=4) n
	integer (kind=4) nb
	integer (kind=4) nunk
    integer (kind=4) xdim
	real (kind=8) x(1:xdim,1:n)
	integer (kind=4) lgnode(1:nb,1:nelements)
	integer (kind=4) unknode(1:n)
    real (kind=8) dt
    real (kind=8) time
	real (kind=8) u(1:n)
	
	! The output error
	real (kind=8) L2Err
	real (kind=8) H1Err
	
	! subroutine specific variables
    integer (kind=4) nquad
	! The local x positions of the nodes
	real (kind=8) lx1(1:xdim)
	real (kind=8) lx2(1:xdim)
	real (kind=8) lx3(1:xdim)
    ! Squared edge lengths of the triangle
    real (kind=8) b
    real (kind=8) c
    real (kind=8) d
	! The quadrature weight
	real (kind=8) w
	! The quadrature point we are looking at
	real (kind=8) xquad_pt(1:xdim,nb)
	! The exact value of the solution and its derivative
	real (kind=8) u_exact
	real (kind=8) dux_exact
	real (kind=8) duy_exact
	! The values of the finite element approx
	real (kind=8) u_fem
	real (kind=8) duy_fem
	real (kind=8) dux_fem
	! Values of the basis functions
	real (kind=8) phii
	real (kind=8) phiix
	real (kind=8) phiiy
	! The values of our functions
	real (kind=8) uu
	real (kind=8) duux
	real (kind=8) duuy
    ! Values used to normalize the errors
    real (kind=8) L2norm
    real (kind=8) H1norm
    	! Counters
	integer (kind=4) iE !elements
	integer (kind=4) iQ !quadrature points
	integer (kind=4) jQ !quadrature points
	integer (kind=4) iB !Basis function i
    
       
	! Initially set the error to zero
	L2Err = 0
	H1Err = 0
    
    ! And the norms
    L2norm = 0.0
    H1norm = 0.0
    
    ! the number of quadrature points
    nquad = 3
	
    ! For all of the elements
    do iE = 1,nelements

        !! Get the size of the interval based on the left
		! and right nodal values
        ! starts in bottom left then goes counter-clockwise
		lx1 = x(:,lgnode(1,iE))
		lx2 = x(:,lgnode(2,iE))
		lx3 = x(:,lgnode(3,iE))
	
        ! Calculate the area of the triangle we are looking at using Herons formula
        ! This is a modification of the formula in order to avoid doing multiple
        ! square roots
        b = ((lx2(2)-lx1(2))**2) + ((lx2(1)-lx1(1))**2)
        c = ((lx3(2)-lx2(2))**2) + ((lx3(1)-lx2(1))**2)
        d = ((lx1(2)-lx3(2))**2) + ((lx1(1)-lx3(1))**2)
        w = (1.0/4.0)*sqrt(abs(((b + c + d)**2) - &
		    (2*((b**2)+(c**2)+(d**2))))) 
        
        ! Find the coordinate of the quadrature point we are on
        xquad_pt(:,1) = (1.0/2)*(lx1+lx2);   
        xquad_pt(:,2) = (1.0/2)*(lx2+lx3);   
        xquad_pt(:,3) = (1.0/2)*(lx3+lx1);
        
        ! The values of the exact u 
        u_exact = 0
	    dux_exact = 0
	    duy_exact = 0
	
        ! For all of the quadrature points (nquad=1)
        do iQ = 1, nquad

            !write(*,'(2x,g14.6)') xquad_pt(:,iQ)
            
            ! Find the exact solution and deriv
            u_exact = uu(xdim,xquad_pt(:,iQ),time)
            dux_exact = duux(xdim,xquad_pt(:,iQ),time)
            duy_exact = duuy(xdim,xquad_pt(:,iQ),time)
            
            !write(*,*) u_exact
           
            ! Calculate the value of the fem at the quad point
            u_fem = 0.0
            dux_fem = 0.0
            duy_fem = 0.0
            
            ! Loop over the basis functions
            do iB = 1, nb
                call phi(xdim, xquad_pt(:,iQ), iB, lx1, lx2, lx3, phii, phiix, phiiy)	
				!write(*,*) phii
                u_fem = u_fem + u(lgnode(iB,iE))*phii
                dux_fem = dux_fem + u(lgnode(iB,iE))*phiix
                duy_fem = duy_fem + u(lgnode(iB,iE))*phiiy
            end do
            
            !write(*,*) u_fem
            !write(*,*)
            
            ! Apply the adjusted weight and add it into the Error
            L2Err = L2Err + (w/3)*((u_exact-u_fem)**2)
            H1Err = H1Err + (w/3)*((u_exact-u_fem)**2) + &
                            (w/3)*((dux_exact-dux_fem)**2) + &
                            (w/3)*((duy_exact-duy_fem)**2)
                            
            ! The norms in order to normalize the error
            L2norm = L2norm + (w/3)*(u_exact**2)
            H1norm = H1norm + (w/3)*(u_exact**2) + &
                              (w/3)*(dux_exact**2) + &
                              (w/3)*(duy_exact**2)

            !write(*,*) dx
            !write(*,*) L2Err
            
            !write(*,*) dux_fem
            !write(*,*) dux_exact
        end do
        
       ! Adjust the error based on the
        
        ! % Now add the squared difference to the error
        ! Err = Err + (u_exact-u_fem).^2;      
    end do
	
	!write(*,*) L2norm
	
	! Finally take the root of both errors
	L2Err = sqrt(L2Err)/sqrt(L2norm)
	H1Err = sqrt(H1Err)/sqrt(H1norm)
    
    !L2Err = sqrt(L2Err)
    !H1Err = sqrt(H1Err)

end

subroutine geometry(xdim, nquad, nb, &
					n, nelements, x, xquad, h, &
					lgnode, nunk, unknode)
	
	! This function defines the grid that we are going to initialize based on
	! our boundary and the discretization values
	
	implicit none
	
	! Define input variables
	integer (kind=4) xdim
	integer (kind=4) nquad
	integer (kind=4) nb
    	
	! Define Output variables
	integer (kind=4) n
	integer (kind=4) nelements
	real (kind=8) x(1:xdim,1:n)
	real (kind=8) xquad(1:xdim,1:nelements)
	real (kind=8) h
	integer (kind=4) lgnode(1:nb,1:nelements)
	integer (kind=4) nunk
	integer (kind=4) unknode(1:n)
	
	! Define Subroutine-specific
	integer (kind=4) i, j
    ! The current point in the unknown array we are looking at
    integer (kind=4) iunk
    ! Keeps track of which element we are on when constructing arrays
    integer (kind=4) ielement
    ! Filenames for our two files we need to read
    character ( len = 80 ) :: XY_csv_filename = 'XY_points.csv'
    character ( len = 80 ) :: NODES_csv_filename = 'Nodes.csv'
    character ( len = 80 ) :: UNK_csv_filename = 'Unknowns.csv'
    character ( len = 80 ) :: H_csv_filename = 'H.txt'
    ! Other variables used to read in the files
    integer   ( kind = 4 ) csv_file_status
    integer   ( kind = 4 ) csv_file_unit
    integer   ( kind = 4 ) csv_record_status

    
    !
	! Read in the xy points array
	!
    
	! Figure out the number of lines, also the number of nodes
	call csv_file_line_count(XY_csv_filename, n)

    ! Open the file to read
    call csv_file_open_read (XY_csv_filename, csv_file_unit)
    
    ! Read the csv file
    do i = 1, n
        read(csv_file_unit, '(g14.0,g14.0)', iostat=csv_file_status) x(:,i)    
    end do
    
    ! Close the XY file
    call csv_file_close_read (XY_csv_filename, csv_file_unit )
	
	
    !
    ! Read in the Nodes array
    !
    
    ! NOTE: The mesh2d function stores nodes of a triangle in counter-clockwise 
    ! orientation so you need to make sure that your basis function routine
    ! also obeys that standard
    
    ! Figure out the number of lines, also the number of nodes
	call csv_file_line_count(NODES_csv_filename, nelements)

	
    ! Open the file to read
    call csv_file_open_read (NODES_csv_filename, csv_file_unit)
    
    ! Read the csv file
    do i = 1, nelements
        read(csv_file_unit, '(i8,i8,i8)', iostat=csv_file_status) lgnode(:,i)    
    end do
    
    ! Close the XY file
    call csv_file_close_read (NODES_csv_filename, csv_file_unit)
    
    !
    ! Now find the quadrature points
    !
    
    ! To calculate a single quadrature point on any given triangle we need
    ! The barycentric coordinates (l1=l2=l3=1/3) dependant on the vertices
    ! of the particular element we are looking at
	do i = 1, nelements

        ! First element top left of the rectangle
        xquad(1,i) = (1.0/3.0)*(x(1,lgnode(1,i)) + &   
                                x(1,lgnode(2,i)) + &    
                                x(1,lgnode(3,i)))
        xquad(2,i) = (1.0/3.0)*(x(2,lgnode(1,i)) + &   
                                x(2,lgnode(2,i)) + &    
                                x(2,lgnode(3,i)))       
        
    end do
	
	!
	! Finally read in the unknown array
	!
    
	! Open the file to read
    call csv_file_open_read (UNK_csv_filename, csv_file_unit)
    
    nunk = 0
    
    ! Read the csv file
    do i = 1, n
        read(csv_file_unit, '(i8)', iostat=csv_file_status) unknode(i)    
        
        ! If there is an unknown increment the unknown number
        if(unknode(i) /= -1) then
            nunk = nunk + 1
        end if        
    end do
    
    ! Close the XY file
    call csv_file_close_read (NODES_csv_filename, csv_file_unit)
	
	!
	! Read in the discretization
	!
	
	! There is only one line in this file

    ! Open the file to read
    call csv_file_open_read (H_csv_filename, csv_file_unit)
    
    ! Read the csv file
    read(csv_file_unit, '(g14.0)', iostat=csv_file_status) h   

    ! Close the file
    call csv_file_close_read (H_csv_filename, csv_file_unit )
	
	
end

subroutine map_x (x,xdim,n)
	! This function is designed to take a uniform grid defined in x
	! and map it to a non-uniform grid, still stored as x
	
	! Inputs:
	! The dimension of x
	integer (kind=4) xdim
    ! The size of x (Also an output)
	integer (kind=4) n
	! The x array
	real (kind=8) x(1:xdim,1:n)
	
	! Routine specific:
	! A counter variable
	integer (kind=4) i
	
	! For all the x values modify them by our function
	! If we want to keep a uniform grid then we just set
	! x(i) = x(i)
	!do i = 1,n
		!x(i) = sqrt(x(i))
		!x(i) = x(i)
	!end do
end


subroutine phi (xdim, x, basis_fn, x1, x2, x3, phii, phiix, phiiy)
	! This function is designed to return the value of our function
	! phi, and its derivative phix at a given point, x
	
	implicit none
	
	! The inputs
	! The value of x we are evaluating
    integer (kind=4) xdim
	real (kind=8) x(1:xdim)
	! The basis function that we are looking at within an element
	integer (kind=4) basis_fn
	! The left and right points of our element
	real (kind=8) x1(1:xdim)
	real (kind=8) x2(1:xdim)
	real (kind=8) x3(1:xdim)
	
	! Outputs
	! The value of phi at a given point
	real (kind=8) phii
	! The value of the derivative of phi in the x and y directions
	real (kind=8) phiix	
	real (kind=8) phiiy	
	
	! Subroutine specific
    ! The normals to each line
    real (kind=8) norm12(1:xdim)
    real (kind=8) norm23(1:xdim)
    real (kind=8) norm31(1:xdim)
    ! The distances to the point from our edges
	real (kind=8) dist12
	real (kind=8) dist23
	real (kind=8) dist31
    
    ! Find the unit normal vectors to each of our edges
    norm12 = (/ (x1(2)-x2(2))/sqrt(((x2(2)-x1(2))**2)+((x2(1)-x1(1))**2)), &
            (x2(1)-x1(1))/sqrt(((x2(2)-x1(2))**2)+((x2(1)-x1(1))**2)) /)
    norm23 = (/ (x2(2)-x3(2))/sqrt(((x3(2)-x2(2))**2)+((x3(1)-x2(1))**2)), &
            (x3(1)-x2(1))/sqrt(((x3(2)-x2(2))**2)+((x3(1)-x2(1))**2)) /)
    norm31 = (/ (x3(2)-x1(2))/sqrt(((x1(2)-x3(2))**2)+((x1(1)-x3(1))**2)), &
            (x1(1)-x3(1))/sqrt(((x1(2)-x3(2))**2)+((x1(1)-x3(1))**2)) /)
    
    ! Find the Signed distances between the points and the edges of our triangle
    dist12 = dot_product(norm12,x-x1)
    dist23 = dot_product(norm23,x-x2)
    dist31 = dot_product(norm31,x-x3)
    
    !write(*,*) dist12
	
	
	! Check to make sure we are in our boundary
    ! anything to the right because we are moving clockwise
	if(dist12 >= -0.00001 .and. dist23 >= -0.00001 .and. dist31 >= -0.00001) then
		
        ! Figure out which basis function we are trying to find
        if(basis_fn == 1) then
            phii = (dist23/dot_product(norm23,x1-x2))
            phiix = (norm23(1)/dot_product(norm23,x1-x2)) 
            phiiy = (norm23(2)/dot_product(norm23,x1-x2))
        elseif(basis_fn == 2) then
            phii = (dist31/dot_product(norm31,x2-x3))
            phiix = (norm31(1)/dot_product(norm31,x2-x3))
            phiiy = (norm31(2)/dot_product(norm31,x2-x3))
        elseif(basis_fn == 3) then
            phii = (dist12/dot_product(norm12,x3-x1))
            phiix = (norm12(1)/dot_product(norm12,x3-x1))
            phiiy = (norm12(2)/dot_product(norm12,x3-x1))
        else
            phii = 0
            phiix = 0
            phiiy = 0
        end if
    
    else
		write(*,*) 'here'
        phii = 0
        phiix = 0
        phiiy = 0
    end if
	
	!if(phii == 0) then
	!	write(*,*) x(:)
	!	write(*,*) x1(:)
	!	write(*,*) x2(:)
	!	write(*,*) x3(:)
	!	write(*,*)
	!end if

end

function ff(xdim, xy, time)
	
	implicit none
	
	! Inputs
    integer (kind=4) xdim
	real (kind=8) xy(1:xdim)
    real (kind=8) time
	
	! Outputs
	real (kind=8) ff
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359
	
	! Evaluate the function at the point
	!ff = (pi*pi)*(sin(pi*x))
	!ff = (2*(pi*pi))*sin(pi*x(1))*sin(pi*x(2))
    !ff = (3*((x(1)**2) - x(1))*((x(2)**2) - x(2))*(time**2)) &
    !     -(2*(time**3)*((x(1)**2) - x(1) + (x(2)**2) - x(2)))
    !ff = (2*xy(1)*((xy(1)**2)-(1.5*xy(1))+0.5)* &
    !     xy(2)*((xy(2)**2)-(1.5*xy(2))+0.5)*(time**2)) - &
    !     ((6*xy(1)-3.0)*xy(2)*((xy(2)**2)-(1.5*xy(2))+0.5)*(time**3))-&
    !     (xy(1)*((xy(1)**2)-(1.5*xy(1))+0.5)*(6*xy(2)-3.0)*(time**3))
    ff = (2.0*time*xy(1)*xy(2)*(1-xy(1)-xy(2))) + &
          (2.0*xy(2)*(time**2)) + (2.0*xy(1)*(time**2))
    
    
	return
end	

function pp(xdim,x)

	implicit none
	
	! Inputs
    integer (kind=4) xdim
	real (kind=8) x(1:xdim)
	
	! Outputs
	real (kind=8) pp
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359
	
	! Evaluate the function at the point
	pp = 1
	
	return
end	

function qq(xdim,x)

	implicit none
	
	! Inputs
    integer (kind=4) xdim
	real (kind=8) x(1:xdim)
	
	! Outputs
	real (kind=8) qq
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359
	
	! Evaluate the function at the point
	qq = 1
	
	return
end

function uu(xdim, xy, time)
	! This is the exact solution 	
	implicit none
	
	! Inputs
    integer (kind=4) xdim
	real (kind=8) xy(1:xdim)
    real (kind=8) time
	
	! Outputs
	real (kind=8) uu
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359
	
	! Evaluate the function at the point
	!uu = sin(pi*x)
	!uu = sin(pi*x(1))*sin(pi*x(2))
	!uu = x(1)*(x(1)-1)*x(2)*(x(2)-1)*(time**3)
    !uu = (xy(1)*((xy(1)**2)-(1.5*xy(1))+0.5)* &
    !      xy(2)*((xy(2)**2)-(1.5*xy(2))+0.5))*(time**3)
    uu = xy(1)*xy(2)*(1-xy(1)-xy(2))*(time**2)
    
	return
end

function duux(xdim, xy, time)
	! This is the derivative of the exact solution
	! used when computing the H1 error norm
	implicit none
	
	! Inputs
    integer (kind=4) xdim
	real (kind=8) xy(1:xdim)
    real (kind=8) time
	
	! Outputs
	real (kind=8) duux
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359
	
	! Evaluate the function at the point
	!duu = pi*cos(pi*x)
	!duux = pi*cos(pi*x(1))*sin(pi*x(2))
	!duux = ((2*x(1)*((x(2)**2)-x(2))) + (x(2)-(x(2)**2)))*(time**3)
	!duux = ((3*(xy(1)**2)-(2.0*xy(1))+0.5)* &
    !        xy(2)*((xy(2)**2)-(1.5*xy(2))+0.5))*(time**3)
    duux = (xy(2)-(xy(2)**2)-(2.0*xy(1)*xy(2)))*(time**2)
	
    return
end

function duuy(xdim, xy, time)
	! This is the derivative of the exact solution
	! used when computing the H1 error norm
	implicit none
	
	! Inputs
    integer (kind=4) xdim
	real (kind=8) xy(1:xdim)
    real (kind=8) time
	
	! Outputs
	real (kind=8) duuy
	
	! routine specific variables
	real (kind=8), parameter :: pi = 3.14159265359
	
	! Evaluate the function at the point
	!duu = pi*cos(pi*x)
	!duuy = pi*sin(pi*x(1))*cos(pi*x(2))
    !duuy = (xy(1)*((xy(1)**2)-(1.5*xy(1))+0.5)* &
     !      (3*(xy(2)**2)-(3.0*xy(2))+0.5))*(time**3)
	duuy = (xy(1)-(xy(1)**2)-(2.0*xy(1)*xy(2)))*(time**2)
    
    
	return
end

